/******************************************************************************
 * File         :   video_elaboration.c
 * Autore       :   Diego Mereghetti, Alessandro Marin
 * Descrizione  :   Applicazione per il riconoscimento della casa base, del
 *                  giocatore e dei waypoint
 * Librerie     :   OpenCV
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <cv.h>
#include <highgui.h>

#define MAX_FILE_PATH               255

#define WINDOW_ORIGINAL             "original"
#define WINDOW_HOME                 "home"
#define WINDOW_PLAYER               "player"
#define WINDOW_WAYPOINT             "waypoint"
#define WINDOW_TEMP_0               "temp0"
#define WINDOW_TEMP_1               "temp1"
#define WINDOW_TEMP_2               "temp2"

#define SUFFIX_HOME                 "_home.mpg"
#define SUFFIX_PLAYER               "_player.mpg"
#define SUFFIX_WAYPOINT             "_waypoint.mpg"

#define DEFAULT_FILE_VIDEO_HOME     "/dev/null"
#define DEFAULT_FILE_VIDEO_PLAYER   "/dev/null"
#define DEFAULT_FILE_VIDEO_WAYPOINT "/dev/null"

#define DEFAULT_FPS                 25

#define DEFAULT_MIN_RG              1.55
#define DEFAULT_MAX_RG              3.02
#define DEFAULT_MIN_GB              0.80
#define DEFAULT_MAX_GB              2.50

#define DEFAULT_HOME_MIN_H          35
#define DEFAULT_HOME_MAX_H          55
#define DEFAULT_HOME_MIN_S          120
#define DEFAULT_HOME_MAX_S          255
#define DEFAULT_HOME_MIN_V          0
#define DEFAULT_HOME_MAX_V          255

#define DEFAULT_PLAYER_MIN_H        2
#define DEFAULT_PLAYER_MAX_H        10
#define DEFAULT_PLAYER_MIN_S        60
#define DEFAULT_PLAYER_MAX_S        255
#define DEFAULT_PLAYER_MIN_V        0
#define DEFAULT_PLAYER_MAX_V        255

#define DEFAULT_WAYPOINT_MIN_Y      255
#define DEFAULT_WAYPOINT_MAX_Y      255
#define DEFAULT_WAYPOINT_MIN_CR     0
#define DEFAULT_WAYPOINT_MAX_CR     255
#define DEFAULT_WAYPOINT_MIN_CB     0
#define DEFAULT_WAYPOINT_MAX_CB     255

#define HOME_HEIGHT_THRESHOLD       0.55
#define WAYPOINT_HEIGHT_THRESHOLD   0.75

#define BW_PIXEL_AT(img, x, y)      ((BWPixel_t*)(img->imageData + y * img->widthStep + x * sizeof(BWPixel_t)))
#define BGR_PIXEL_AT(img, x, y)     ((BGRPixel_t*)(img->imageData + y * img->widthStep + x * sizeof(BGRPixel_t)))
#define HSV_PIXEL_AT(img, x, y)     ((HSVPixel_t*)(img->imageData + y * img->widthStep + x * sizeof(HSVPixel_t)))
#define YCrCb_PIXEL_AT(img, x, y)   ((YCrCbPixel_t*)(img->imageData + y * img->widthStep + x * sizeof(YCrCbPixel_t)))


typedef unsigned char BWPixel_t;

typedef struct
{
    unsigned char b;
    unsigned char g;
    unsigned char r;
} BGRPixel_t;

typedef struct
{
    unsigned char h;
    unsigned char s;
    unsigned char v;
} HSVPixel_t;

typedef struct
{
    unsigned char y;
    unsigned char cr;
    unsigned char cb;
} YCrCbPixel_t;


int BGR_pixel_recognized(BGRPixel_t* pixel, double min_rg, double max_rg, double min_gb, double max_gb);
int HSV_pixel_reconized(HSVPixel_t* pixel, double min_h, double max_h, double min_s, double max_s, double min_v, double max_v);
int YCrCb_pixel_reconized(YCrCbPixel_t* pixel, double min_y, double max_y, double min_cr, double max_cr, double min_cb, double max_cb);


int main(int argc, const char* argv[])
{
    CvCapture*      capture;
    CvVideoWriter*  writer_home;
    CvVideoWriter*  writer_player;
    CvVideoWriter*  writer_waypoint;
    char            video_home[MAX_FILE_PATH + 1];
    char            video_player[MAX_FILE_PATH + 1];
    char            video_waypoint[MAX_FILE_PATH + 1];
    IplImage*       frame;
    IplImage*       original_frame_B;
    IplImage*       original_frame_G;
    IplImage*       original_frame_R;
    IplImage*       original_frame_hsv;
    IplImage*       original_frame_ycrcb;
    IplImage*       filtered_frame_home;
    IplImage*       filtered_frame_player;
    IplImage*       filtered_frame_waypoint;
    IplImage*       final_frame_home;
    IplImage*       final_frame_player;
    IplImage*       final_frame_waypoint;
    double          fps;
    double          min_rg, max_rg, min_gb, max_gb;
    int             min_h, max_h, min_s, max_s, min_v, max_v;
    int             min_y, max_y, min_cr, max_cr, min_cb, max_cb;
    int             x, y;
    
    
    if ( argc < 2 )
    {
        fprintf(stderr, "USAGE: %s <videofile> [<outvideo_prefix> <fps>]\n", argv[0]);
        exit(-1);
    }   
    else if ( argc < 4 )
    {
        strncpy(video_home,     DEFAULT_FILE_VIDEO_HOME,        MAX_FILE_PATH + 1);
        strncpy(video_player,   DEFAULT_FILE_VIDEO_PLAYER,      MAX_FILE_PATH + 1);
        strncpy(video_waypoint, DEFAULT_FILE_VIDEO_WAYPOINT,    MAX_FILE_PATH + 1);
        fps = DEFAULT_FPS;
    }
    else
    {
        strncpy(video_home,     argv[2],    MAX_FILE_PATH + 1); strncat(video_home,     SUFFIX_HOME,        MAX_FILE_PATH - strlen(video_home));
        strncpy(video_player,   argv[2],    MAX_FILE_PATH + 1); strncat(video_player,   SUFFIX_PLAYER,      MAX_FILE_PATH - strlen(video_player));
        strncpy(video_waypoint, argv[2],    MAX_FILE_PATH + 1); strncat(video_waypoint, SUFFIX_WAYPOINT,    MAX_FILE_PATH - strlen(video_waypoint));
        fps = atof(argv[3]);
    }
    
    capture = cvCreateFileCapture(argv[1]);
    if ( capture == NULL )
    {
        fprintf(stderr, "file not existing or non supported video codec\n");
        exit(-1);
    }
    
    
    // cvNamedWindow(WINDOW_ORIGINAL,  CV_WINDOW_AUTOSIZE);
    // cvNamedWindow(WINDOW_HOME,      CV_WINDOW_AUTOSIZE);
    // cvNamedWindow(WINDOW_PLAYER,    CV_WINDOW_AUTOSIZE);
    // cvNamedWindow(WINDOW_WAYPOINT,  CV_WINDOW_AUTOSIZE);
    
    
    // Read first frame to obtain width & height
    frame = cvQueryFrame(capture);
    if ( frame == NULL )
    {
        fprintf(stderr, "Unable to query frame\n");
        cvReleaseCapture(&capture);
        exit(-1);
    }
    
    // Prepare video writers
    // writer_home     = cvCreateVideoWriter(video_home,       CV_FOURCC('M', 'J', 'P', 'G'), fps, cvGetSize(frame), 1);
    // writer_player   = cvCreateVideoWriter(video_player,     CV_FOURCC('M', 'J', 'P', 'G'), fps, cvGetSize(frame), 1);
    // writer_waypoint = cvCreateVideoWriter(video_waypoint,   CV_FOURCC('M', 'J', 'P', 'G'), fps, cvGetSize(frame), 1);
    
    // writer_home     = cvCreateVideoWriter(video_home,       CV_FOURCC('M', '4', 'S', '2'), fps, cvGetSize(frame), 1);
    // writer_player   = cvCreateVideoWriter(video_player,     CV_FOURCC('M', '4', 'S', '2'), fps, cvGetSize(frame), 1);
    // writer_waypoint = cvCreateVideoWriter(video_waypoint,   CV_FOURCC('M', '4', 'S', '2'), fps, cvGetSize(frame), 1);
    
    writer_home     = cvCreateVideoWriter(video_home,       CV_FOURCC('P', 'I', 'M', '1'), fps, cvGetSize(frame), 1);
    writer_player   = cvCreateVideoWriter(video_player,     CV_FOURCC('P', 'I', 'M', '1'), fps, cvGetSize(frame), 1);
    writer_waypoint = cvCreateVideoWriter(video_waypoint,   CV_FOURCC('P', 'I', 'M', '1'), fps, cvGetSize(frame), 1);
    
    
    // Prepare helper layers
    original_frame_B        = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    original_frame_G        = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    original_frame_R        = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    original_frame_hsv      = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    original_frame_ycrcb    = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    filtered_frame_home     = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    filtered_frame_player   = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    filtered_frame_waypoint = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    final_frame_home        = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    final_frame_player      = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    final_frame_waypoint    = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    
    
    do
    {
        frame = cvQueryFrame(capture);
        if ( frame == NULL ) break;
        
        cvCvtColor(frame, original_frame_hsv,   CV_BGR2HSV);
        cvCvtColor(frame, original_frame_ycrcb, CV_BGR2YCrCb);
        
        // Player identification ------------------------------------------------------------------
        // Do the identification in HSV colorspace
        for ( y = 0; y < original_frame_hsv->height; y++ )
        {
            for ( x = 0; x < original_frame_hsv->width; x++ )
            {
                if ( HSV_pixel_reconized(HSV_PIXEL_AT(original_frame_hsv, x, y),
                    DEFAULT_PLAYER_MIN_H, DEFAULT_PLAYER_MAX_H,
                    DEFAULT_PLAYER_MIN_S, DEFAULT_PLAYER_MAX_S,
                    DEFAULT_PLAYER_MIN_S, DEFAULT_PLAYER_MAX_V) )                           *(BW_PIXEL_AT(filtered_frame_player, x, y)) = 255;
                else                                                                        *(BW_PIXEL_AT(filtered_frame_player, x, y)) = 0;
            }
        }
        
        
        // Home identification ------------------------------------------------------------------
        for ( y = 0; y < original_frame_hsv->height; y++ )
        {
            for ( x = 0; x < original_frame_hsv->width; x++ )
            {
                if ( y <= HOME_HEIGHT_THRESHOLD * original_frame_hsv->height )              *(BW_PIXEL_AT(filtered_frame_home, x, y)) = 0;
                else if ( HSV_pixel_reconized(HSV_PIXEL_AT(original_frame_hsv, x, y),
                    DEFAULT_HOME_MIN_H, DEFAULT_HOME_MAX_H,
                    DEFAULT_HOME_MIN_S, DEFAULT_HOME_MAX_S,
                    DEFAULT_HOME_MIN_V, DEFAULT_HOME_MAX_V) )                               *(BW_PIXEL_AT(filtered_frame_home, x, y)) = 255;
                else                                                                        *(BW_PIXEL_AT(filtered_frame_home, x, y)) = 0;
            }
        }
        
        
        // Waypoint identification --------------------------------------------------------------
        for ( y = 0; y < original_frame_ycrcb->height; y++ )
        {
            for ( x = 0; x < original_frame_ycrcb->width; x++ )
            {
                if ( y <= WAYPOINT_HEIGHT_THRESHOLD * original_frame_ycrcb->height )        *(BW_PIXEL_AT(filtered_frame_waypoint, x, y)) = 0;
                else if ( YCrCb_pixel_reconized(YCrCb_PIXEL_AT(original_frame_ycrcb, x, y),
                    DEFAULT_WAYPOINT_MIN_Y,  DEFAULT_WAYPOINT_MAX_Y,
                    DEFAULT_WAYPOINT_MIN_CR, DEFAULT_WAYPOINT_MAX_CR,
                    DEFAULT_WAYPOINT_MIN_CB, DEFAULT_WAYPOINT_MAX_CB) )                     *(BW_PIXEL_AT(filtered_frame_waypoint, x, y)) = 255;
                else                                                                        *(BW_PIXEL_AT(filtered_frame_waypoint, x, y)) = 0;
            }
        }
        
        cvCvtColor(filtered_frame_home,     final_frame_home,       CV_GRAY2BGR);
        cvCvtColor(filtered_frame_player,   final_frame_player,     CV_GRAY2BGR);
        cvCvtColor(filtered_frame_waypoint, final_frame_waypoint,   CV_GRAY2BGR);
        
        // cvShowImage(WINDOW_ORIGINAL,    frame);
        // cvShowImage(WINDOW_HOME,        final_frame_home);
        // cvShowImage(WINDOW_PLAYER,      final_frame_player);
        // cvShowImage(WINDOW_WAYPOINT,    final_frame_waypoint);
        
        cvWriteFrame(writer_home,       final_frame_home);
        cvWriteFrame(writer_player,     final_frame_player);
        cvWriteFrame(writer_waypoint,   final_frame_waypoint);
        
    // } while ( cvWaitKey(100) != 'q' );
    } while ( 1 );
    
    cvReleaseVideoWriter(&writer_home);
    cvReleaseVideoWriter(&writer_player);
    cvReleaseVideoWriter(&writer_waypoint);
    
    cvReleaseImage(&original_frame_hsv);
    cvReleaseImage(&filtered_frame_home);
    cvReleaseImage(&filtered_frame_player);
    cvReleaseImage(&filtered_frame_waypoint);
    cvReleaseImage(&final_frame_home);
    cvReleaseImage(&final_frame_player);
    cvReleaseImage(&final_frame_waypoint);
    
    cvReleaseCapture(&capture);
    
    
    printf("Done\n");
    
    return 0;
}


int BGR_pixel_recognized(BGRPixel_t* pixel, double min_rg, double max_rg, double min_gb, double max_gb)
{   
    if (
        ( (double)(pixel->r)/(double)(pixel->g) <= max_rg ) &&
        ( (double)(pixel->r)/(double)(pixel->g) >= min_rg ) &&
        ( (double)(pixel->g)/(double)(pixel->b) <= max_gb ) &&
        ( (double)(pixel->g)/(double)(pixel->b) >= min_gb ) ) return 1;
    else return 0;
}


int HSV_pixel_reconized(HSVPixel_t* pixel, double min_h, double max_h, double min_s, double max_s, double min_v, double max_v)
{
    if (
        (pixel->h >= min_h ) && (pixel->h <= max_h ) &&
        (pixel->s >= min_s ) && (pixel->s <= max_s ) &&
        (pixel->v >= min_v ) && (pixel->v <= max_v ) ) return 1;
    else return 0;
}


int YCrCb_pixel_reconized(YCrCbPixel_t* pixel, double min_y, double max_y, double min_cr, double max_cr, double min_cb, double max_cb)
{
    if (
        (pixel->y  >= min_y  ) && (pixel->y  <= max_y  ) &&
        (pixel->cr >= min_cr ) && (pixel->cr <= max_cr ) &&
        (pixel->cb >= min_cb ) && (pixel->cb <= max_cb ) ) return 1;
    else return 0;
}
