/******************************************************************************
 * File         :   wizard_of_oz.c
 * Autore       :   Diego Mereghetti, Alessandro Marin
 * Descrizione  :   Applicazione per la simulazione del gioco
 * Librerie     :   libcwiid
 *                  libbluetooth
 * Note         :   libwiimote è stato sostituito da libcwiid in quanto più
 *                  stabile
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include <bluetooth/bluetooth.h>
#include <cwiid.h>


#define WEAPON_PISTOL           0
#define WEAPON_RIFLE            1
#define WEAPON_SWORD            2

#define MULTIPLIER_PISTOL       1
#define MULTIPLIER_RIFLE        2

#define MAX_PISTOL_BULLETS      4
#define MAX_RIFLE_BULLETS       2
#define MAX_SWORD_BULLETS       0

#define MAX_LEDS_YELLOW         4
#define MAX_LEDS_RED            4

#define LED_GREEN               0x08
#define LEDS_RED_0              0x00
#define LEDS_RED_1              0x01
#define LEDS_RED_2              0x02
#define LEDS_RED_3              0x03
#define LEDS_RED_4              0x04
#define LED_YELLOW_1            0x10
#define LED_YELLOW_2            0x20
#define LED_YELLOW_3            0x40
#define LED_YELLOW_4            0x80

#define SHOW_DELAY              30000
#define SHOW_REPEATS            25

#define DAMAGE_4_LED            1

#define DEFAULT_ROBOT_STAMINA   100
#define DEFAULT_WEAPON          WEAPON_PISTOL
#define DEFAULT_BULLETS         MAX_PISTOL_BULLETS

#define DELAY                   20000
#define TARGETING_CYCLES        (800000/DELAY)
#define RUMBLE_DELAY            400000

#define FILTER_VALUE_A          (350000/DELAY)
#define FILTER_VALUE_B          (350000/DELAY)
#define FILTER_VALUE_ONE        (300000/DELAY)
#define FILTER_VALUE_TWO        (300000/DELAY)
#define FILTER_VALUE_MINUS      (40000/DELAY)
#define FILTER_VALUE_PLUS       (40000/DELAY)
#define FILTER_VALUE_HOME       (40000/DELAY)
#define FILTER_VALUE_UP         (100000/DELAY)
#define FILTER_VALUE_DOWN       (100000/DELAY)
#define FILTER_VALUE_LEFT       (100000/DELAY)
#define FILTER_VALUE_RIGHT      (100000/DELAY)


typedef struct
{
    uint8_t a;
    uint8_t b;
    uint8_t one;
    uint8_t two;
    uint8_t minus;
    uint8_t plus;
    uint8_t home;
    uint8_t up;
    uint8_t down;
    uint8_t left;
    uint8_t right;
} wiimote_status_t;


typedef struct
{
    uint8_t filter_a;
    uint8_t filter_b;
    uint8_t filter_one;
    uint8_t filter_two;
    uint8_t filter_minus;
    uint8_t filter_plus;
    uint8_t filter_home;
    uint8_t filter_up;
    uint8_t filter_down;
    uint8_t filter_left;
    uint8_t filter_right;
} filter_t;


typedef struct
{
    uint16_t    robot_stamina;
    uint8_t     led_green;
    uint8_t     leds_red;
    uint8_t     leds_yellow;
    uint16_t    pointing_the_player;
    uint16_t    pointing_the_robot;
    uint8_t     current_weapon;
    uint8_t     bullets;
} game_status_t;


void next_weapon(game_status_t* game_status);
void prev_weapon(game_status_t* game_status);
void reload_weapon(game_status_t* game_status);
void shot(game_status_t* game_status);

void rumble(cwiid_wiimote_t* wiimote, useconds_t timeout);
void show1(FILE* robot_interface, game_status_t* game_status);

void target_lost(game_status_t* game_status);
void targeting(game_status_t* game_status);
void too_closer(game_status_t* game_status);

void reset_game(game_status_t* game_status, filter_t* filter);
void state_wrapper(wiimote_status_t* wiimote_status, filter_t* filter, struct cwiid_state* state);
void update_robot(FILE* robot_interface, game_status_t* game_status);
void update_wiimote(cwiid_wiimote_t* wiimote, game_status_t* game_status);


int main(int argc, const char* argv[])
{
    cwiid_wiimote_t*    wiimote;
    struct cwiid_state  state;
    wiimote_status_t    wiimote_status;
    bdaddr_t            addr;
    filter_t            filter;
    game_status_t       game_status;
    FILE*               robot;
    
    
    if ( argc < 3 )
    {
        fprintf(stderr, "USAGE: %s <base_address> <robot_interface>\n", argv[0]);
        exit(-1);
    }
    
    
    // Connect to wiimote
    str2ba(argv[1], &addr);
    wiimote = cwiid_open(&addr, 0);
    cwiid_command(wiimote, CWIID_CMD_RPT_MODE, CWIID_RPT_STATUS | CWIID_RPT_BTN | CWIID_RPT_ACC | CWIID_RPT_IR);
    
    
    // Connect to robot
    robot = fopen(argv[2], "a");
    if ( robot == NULL )
    {
        perror("fopen");
        cwiid_close(wiimote);
        exit(-1);
    }
    
    
    // Reset game
    reset_game(&game_status, &filter);
    
    
    // Main loop
    while (1)
    {
        cwiid_get_state(wiimote, &state);
        state_wrapper(&wiimote_status, &filter, &state);
        
        printf("weapon=%d bullets=%d ", game_status.current_weapon, game_status.bullets);
        printf("x=%u y=%u z=%u ", state.acc[0], state.acc[1], state.acc[2]);
        printf("ir=%d ", state.ir_src[0].valid + state.ir_src[1].valid + state.ir_src[2].valid + state.ir_src[3].valid);
        printf("targeting=%d ", game_status.pointing_the_robot);
        printf("yellow=%u red=%u green=%u ", game_status.leds_yellow, game_status.leds_red, game_status.led_green);
        printf("\r");
        fflush(stdout);
        
        switch( state.ir_src[0].valid + state.ir_src[1].valid + state.ir_src[2].valid + state.ir_src[3].valid + state.ir_src[4].valid )
        {
            case 4:
            too_closer(&game_status);
            break;
            
            case 3:
            case 2:
            targeting(&game_status);
            break;
            
            case 1:
            case 0:
            target_lost(&game_status);
            break;
        }
        
        if ( wiimote_status.a       ) reload_weapon(&game_status);
        if ( wiimote_status.b       ) shot(&game_status);
        if ( wiimote_status.one     ) reset_game(&game_status, &filter);
        if ( wiimote_status.two     ) show1(robot, &game_status);
        if ( wiimote_status.left    ) { prev_weapon(&game_status);  rumble(wiimote, RUMBLE_DELAY);   }
        if ( wiimote_status.right   ) { next_weapon(&game_status);  rumble(wiimote, RUMBLE_DELAY);   }
        if ( wiimote_status.home    ) break;
        
        update_wiimote(wiimote, &game_status);
        update_robot(robot, &game_status);
        usleep(DELAY);
    }
    
    fclose(robot);
    cwiid_close(wiimote);
    
    return 0;
}


void rumble(cwiid_wiimote_t* wiimote, useconds_t timeout)
{
    cwiid_command(wiimote, CWIID_CMD_RUMBLE, 1);
    usleep(timeout);
    cwiid_command(wiimote, CWIID_CMD_RUMBLE, 0);
}


void show1(FILE* robot_interface, game_status_t* game_status)
{
    uint8_t to_be_sent;
    int     i;
    
    for ( i = 0; i < SHOW_REPEATS; i++ )
    {
        to_be_sent = LEDS_RED_0;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LEDS_RED_1;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LED_YELLOW_3 | LEDS_RED_2;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LED_YELLOW_3 | LED_YELLOW_2 | LEDS_RED_3;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LED_YELLOW_3 | LED_YELLOW_2  | LED_YELLOW_1 | LEDS_RED_4;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LED_YELLOW_3 | LED_YELLOW_2 | LEDS_RED_3;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LED_YELLOW_3 | LEDS_RED_2;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LED_YELLOW_4 | LEDS_RED_1;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
        
        to_be_sent = LEDS_RED_0;
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        usleep(SHOW_DELAY);
    }
    
    game_status->led_green      = 0;
    game_status->leds_red       = 0;
    game_status->leds_yellow    = 0;
}


void next_weapon(game_status_t* game_status)
{
    switch ( game_status->current_weapon )
    {
        case WEAPON_PISTOL  : game_status->current_weapon = WEAPON_RIFLE;   game_status->bullets = MAX_RIFLE_BULLETS;   break;
        case WEAPON_RIFLE   : game_status->current_weapon = WEAPON_SWORD;   game_status->bullets = MAX_SWORD_BULLETS;   break;
        case WEAPON_SWORD   : game_status->current_weapon = WEAPON_PISTOL;  game_status->bullets = MAX_PISTOL_BULLETS;  break;
        default             : game_status->current_weapon = WEAPON_PISTOL;  game_status->bullets = MAX_PISTOL_BULLETS;  break;
    }
}


void prev_weapon(game_status_t* game_status)
{
    switch ( game_status->current_weapon )
    {
        case WEAPON_PISTOL  : game_status->current_weapon = WEAPON_SWORD;   game_status->bullets = MAX_SWORD_BULLETS;   break;
        case WEAPON_RIFLE   : game_status->current_weapon = WEAPON_PISTOL;  game_status->bullets = MAX_PISTOL_BULLETS;  break;
        case WEAPON_SWORD   : game_status->current_weapon = WEAPON_RIFLE;   game_status->bullets = MAX_RIFLE_BULLETS;   break;
        default             : game_status->current_weapon = WEAPON_PISTOL;  game_status->bullets = MAX_PISTOL_BULLETS;  break;
    }
}


void reload_weapon(game_status_t* game_status)
{
    switch ( game_status->current_weapon )
    {
        case WEAPON_PISTOL  : game_status->bullets = MAX_PISTOL_BULLETS;    break;
        case WEAPON_RIFLE   : game_status->bullets = MAX_RIFLE_BULLETS;     break;
        case WEAPON_SWORD   : game_status->bullets = MAX_SWORD_BULLETS;     break;
    }
}


void shot(game_status_t* game_status)
{
    switch ( game_status->current_weapon )
    {
        case WEAPON_PISTOL:
        if ( game_status->bullets <= 0 ) return;
        game_status->bullets--;
        switch ( game_status->leds_yellow )
        {
            case 0:
            break;
            
            case 1:
            case 2:
            case 3:
            // TODO stamina
            // game_status->robot_stamina -= 1
            break;
            
            case 4:
            game_status->leds_red += MULTIPLIER_PISTOL * DAMAGE_4_LED;
            break;
        }
        break;
        
        case WEAPON_RIFLE:
        if ( game_status->bullets <= 0 ) return;
        game_status->bullets--;
        switch ( game_status->leds_yellow )
        {
            case 0:
            break;
            
            case 1:
            case 2:
            case 3:
            // TODO stamina
            // game_status->robot_stamina -= 1
            break;
            
            case 4:
            game_status->leds_red += MULTIPLIER_RIFLE * DAMAGE_4_LED;
            break;
        }
        break;
        
        case WEAPON_SWORD:
        // TODO stamina
        break;
    }
    
    game_status->pointing_the_robot = 0;
    game_status->leds_yellow        = 0;
    game_status->leds_red           = ( game_status->leds_red <= MAX_LEDS_RED ) ? game_status->leds_red : MAX_LEDS_RED;
}


void target_lost(game_status_t* game_status)
{
    game_status->pointing_the_robot = 0;
    game_status->led_green          = 0;
    game_status->leds_yellow        = 0;
}


void targeting(game_status_t* game_status)
{
    game_status->pointing_the_robot++;
    game_status->led_green          = 0;
    game_status->leds_yellow        = ((double)game_status->pointing_the_robot/(double)TARGETING_CYCLES <= MAX_LEDS_YELLOW) ? (double)game_status->pointing_the_robot/(double)TARGETING_CYCLES : MAX_LEDS_YELLOW;
}


void too_closer(game_status_t* game_status)
{
    game_status->pointing_the_robot = 0;
    game_status->led_green          = 1;
    game_status->leds_yellow        = 0;
}


void reset_game(game_status_t* game_status, filter_t* filter)
{
    game_status->robot_stamina          = DEFAULT_ROBOT_STAMINA;
    game_status->pointing_the_player    = 0;
    game_status->pointing_the_robot     = 0;
    game_status->led_green              = 0;
    game_status->leds_red               = 0;
    game_status->leds_yellow            = 0;
    game_status->current_weapon         = DEFAULT_WEAPON;
    game_status->bullets                = DEFAULT_BULLETS;
    filter->filter_a                    = 0;
    filter->filter_b                    = 0;
    filter->filter_one                  = 0;
    filter->filter_two                  = 0;
    filter->filter_minus                = 0;
    filter->filter_plus                 = 0;
    filter->filter_home                 = 0;
    filter->filter_up                   = 0;
    filter->filter_down                 = 0;
    filter->filter_left                 = 0;
    filter->filter_right                = 0;
}


void state_wrapper(wiimote_status_t* wiimote_status, filter_t* filter, struct cwiid_state* state)
{
    wiimote_status->a       = 0;
    wiimote_status->b       = 0;
    wiimote_status->one     = 0;
    wiimote_status->two     = 0;
    wiimote_status->minus   = 0;
    wiimote_status->plus    = 0;
    wiimote_status->home    = 0;
    wiimote_status->up      = 0;
    wiimote_status->down    = 0;
    wiimote_status->left    = 0;
    wiimote_status->right   = 0;
    
    if ( filter->filter_a )     filter->filter_a--;
    if ( filter->filter_b )     filter->filter_b--;
    if ( filter->filter_one )   filter->filter_one--;
    if ( filter->filter_two )   filter->filter_two--;
    if ( filter->filter_minus ) filter->filter_minus--;
    if ( filter->filter_plus )  filter->filter_plus--;
    if ( filter->filter_home )  filter->filter_home--;
    if ( filter->filter_up )    filter->filter_up--;
    if ( filter->filter_down )  filter->filter_down--;
    if ( filter->filter_left )  filter->filter_left--;
    if ( filter->filter_right ) filter->filter_right--;
    
    if ( state->buttons & CWIID_BTN_A     && !(filter->filter_a)        )   { wiimote_status->a     = 1;    filter->filter_a        = FILTER_VALUE_A;       }
    if ( state->buttons & CWIID_BTN_B     && !(filter->filter_b)        )   { wiimote_status->b     = 1;    filter->filter_b        = FILTER_VALUE_B;       }
    if ( state->buttons & CWIID_BTN_1     && !(filter->filter_one)      )   { wiimote_status->one   = 1;    filter->filter_one      = FILTER_VALUE_ONE;     }
    if ( state->buttons & CWIID_BTN_2     && !(filter->filter_two)      )   { wiimote_status->two   = 1;    filter->filter_two      = FILTER_VALUE_TWO;     }
    if ( state->buttons & CWIID_BTN_MINUS && !(filter->filter_minus)    )   { wiimote_status->minus = 1;    filter->filter_minus    = FILTER_VALUE_MINUS;   }
    if ( state->buttons & CWIID_BTN_PLUS  && !(filter->filter_plus)     )   { wiimote_status->plus  = 1;    filter->filter_plus     = FILTER_VALUE_PLUS;    }
    if ( state->buttons & CWIID_BTN_HOME  && !(filter->filter_home)     )   { wiimote_status->home  = 1;    filter->filter_home     = FILTER_VALUE_HOME;    }
    if ( state->buttons & CWIID_BTN_UP    && !(filter->filter_up)       )   { wiimote_status->up    = 1;    filter->filter_up       = FILTER_VALUE_UP;      }
    if ( state->buttons & CWIID_BTN_DOWN  && !(filter->filter_down)     )   { wiimote_status->down  = 1;    filter->filter_down     = FILTER_VALUE_DOWN;    }
    if ( state->buttons & CWIID_BTN_LEFT  && !(filter->filter_left)     )   { wiimote_status->left  = 1;    filter->filter_left     = FILTER_VALUE_LEFT;    }
    if ( state->buttons & CWIID_BTN_RIGHT && !(filter->filter_right)    )   { wiimote_status->right = 1;    filter->filter_right    = FILTER_VALUE_RIGHT;   }
}


void update_robot(FILE* robot_interface, game_status_t* game_status)
{
    static uint8_t  to_be_sent_old  = 1;
    uint8_t         to_be_sent      = 0;
    
    
    // Green led
    to_be_sent |= (game_status->led_green) ? LED_GREEN : 0;
    
    // Red leds
    switch ( game_status->leds_red )
    {
        case 4:     to_be_sent |= LEDS_RED_4;   break;
        case 3:     to_be_sent |= LEDS_RED_3;   break;
        case 2:     to_be_sent |= LEDS_RED_2;   break;
        case 1:     to_be_sent |= LEDS_RED_1;   break;
        default:    break;
    }
    
    // Yellow leds
    switch ( game_status->leds_yellow )
    {
        case 4:     to_be_sent |= LED_YELLOW_4; // NO break
        case 3:     to_be_sent |= LED_YELLOW_3;
        case 2:     to_be_sent |= LED_YELLOW_2;
        case 1:     to_be_sent |= LED_YELLOW_1;
        default:    break;
    }
    
    /*
    int i;
    for( i = 7; i >= 0; i-- )
    {
        if ( ( to_be_sent >> i ) & 0x01 ) printf("1");
        else printf("0");
        
        if ( ( i == 4 ) || ( i == 3 ) ) printf(" ");
    }
    printf("\n");
    */
    
    if ( to_be_sent_old == to_be_sent ) return;
    else
    {
        fwrite(&to_be_sent, 1, 1, robot_interface);
        fflush(robot_interface);
        to_be_sent_old = to_be_sent;
    }
}


void update_wiimote(cwiid_wiimote_t* wiimote, game_status_t* game_status)
{
    static int  leds_flag_old   = 1;
    int         leds_flag       = 0;
    
    switch ( game_status->bullets )
    {
        case 4:
        leds_flag |= CWIID_LED4_ON;
        
        case 3:
        leds_flag |= CWIID_LED3_ON;
        
        case 2:
        leds_flag |= CWIID_LED2_ON;
        
        case 1:
        leds_flag |= CWIID_LED1_ON;
        break;
    }
    
    if ( leds_flag_old == leds_flag ) return;
    else
    {
        cwiid_command(wiimote, CWIID_CMD_LED, leds_flag);
        leds_flag_old = leds_flag;
    }
}
